import math
import datetime


class myMoon:
    def __init__(self):
        self.lunartime = 29.530588853

    def get_julian_date(self):
        date = None
        # If no date is provided, use the current date and time
        if date is None:
            date = datetime.datetime.now()

        # Convert the date to a timestamp in milliseconds
        timestamp = date.timestamp() * 1000

        # Calculate the timezone offset in minutes
        if date.utcoffset():
            timezone_offset = date.utcoffset().total_seconds() // 60
        else:
            timezone_offset = 0

        # Calculate the Julian date and return it
        julian_date = (timestamp / 86400000) - (timezone_offset / 1440) + 2440587.5
        return julian_date

    def get_lunar_age(self):
        percent = self.get_lunar_age_percent()
        age = percent * self.lunartime
        return age

    def get_lunar_age_percent(self):
        julian_date = self.get_julian_date()
        tmp = (julian_date - 2451550.1) / self.lunartime
        percent = self.normalize(tmp)
        return percent

    def normalize(self, value):
        value = value - int(value)
        if value < 0:
            value = value + 1
        return value
    
    def moon_illumination_percent(self, phase):
        
        """
        Calculates the illumination percent of the moon based on its phase.

        Parameters:
            phase (float): The phase of the moon, expressed as a value between 0 and 1.

        Returns:
            float: The percentage of the moon's visible surface that is illuminated, expressed as a value between 0 and 100.
        """

        # Convert phase to angle in radians
        angle = 2 * math.pi * phase

        # Calculate the fraction of the moon's visible surface that is illuminated
        fraction = (1 + math.cos(angle)) / 2

        # Convert fraction to percentage
        percent = fraction * 100

        return percent


    def get_lunar_phase(self):
        age = self.get_lunar_age()
        if age < 1.84566:
            return "new moon"
        elif age < 5.53699:
            return "waxing crescent"
        elif age < 9.22831:
            return "first quarter"
        elif age < 12.91963:
            return "waxing gibbous"
        elif age < 16.61096:
            return "full"
        elif age < 20.30228:
            return 'waning gibbous'
        elif age < 23.99361:
            return "third quarter"
        elif age < 27.68493:
            return "waning crescent"

        # in case it has just finished it's cycle
        return "new"

    def get_lunar_phase_description(self):
        age = self.get_lunar_age()
        if age < 1.84566:
            return "new moon: the marking of the new beginning of the lunar cycle. The new moon shows up around once a month (every 29.5 days). During this time, the moon is in line with the sun, causing its illumination to be around 0%."
        elif age < 5.53699:
            return "waxing crescent: the second phase in the cycle of phases. This Moon phase occurs once a month, rising around 9 AM, and setting around 9 PM, sticking around for approximately 7.38 days before going into the First Quarter phase. "
        elif age < 9.22831:
            return "first quarter: rises around noon and sets around midnight. It’s high in the sky in the evening and makes for excellent viewing."
        elif age < 12.91963:
            return "waxing gibbous: when the lit-up part of the Moon grows from 50.1% to 99.9%. It starts just after the First Quarter Moon and lasts until the Full Moon."
        elif age < 16.61096:
            return "full: when the Sun and the Moon are aligned on opposite sides of Earth, and 100% of the Moon's face is illuminated by the Sun"
        elif age < 20.30228:
            return 'waning gibbous: when the lit-up part of the Moon shrinks from 99.9% to 50.1%. It starts just after Full Moon and lasts until the Third Quarter Moon.'
        elif age < 23.99361:
            return "third quarter: the seventh phase in the cycle of phases. This Moon phase occurs once a month, rising around 12 AM, and setting around 12 PM, almost instantaneously becoming a Waning Crescent."
        elif age < 27.68493:
            return "waning crescent: the eighth and final phase in the cycle of phases. This Moon phase occurs once a month, rising around 3 AM, and setting around 3 PM, sticking around for approximately 7.38 days before going into the New Moon phase."

        return "new moon: the marking of the new beginning of the lunar cycle. The new moon shows up around once a month (every 29.5 days). During this time, the moon is in line with the sun, causing its illumination to be around 0%."

    def moon_alt_az(self, lat, lon, date):
        # Convert latitude and longitude to radians
        lat = math.radians(lat)
        lon = math.radians(lon)

        # Calculate the Julian date
        J2000 = 2451545
        J = date.timestamp() / 86400 + 2440587.5 - J2000

        # Calculate the moon's position in radians
        N = math.radians((125.1228 - 0.0529538083 * J) % 360)
        i = math.radians(5.1454)
        w = math.radians((318.0634 + 0.1643573223 * J) % 360)
        a = 60.2666
        e = 0.054900
        M = math.radians((115.3654 + 13.0649929509 * J) % 360)
        E = M + e * math.sin(M) * (1.0 + e * math.cos(M))
        xv = a * (math.cos(E) - e)
        yv = a * (math.sqrt(1.0 - e * e) * math.sin(E))
        v = math.atan2(yv, xv)
        r = math.sqrt(xv * xv + yv * yv)
        xh = r * (math.cos(N) * math.cos(v + w) - math.sin(N) * math.sin(v + w) * math.cos(i))
        yh = r * (math.sin(N) * math.cos(v + w) + math.cos(N) * math.sin(v + w) * math.cos(i))
        zh = r * (math.sin(v + w) * math.sin(i))

        # Calculate the Greenwich sidereal time
        JD = date.timestamp() / 86400 + 2440587.5
        T = (JD - 2451545.0) / 36525
        L0 = math.radians(280.4665 + 36000.7698 * T)
        dL = math.radians(218.3165 + 481267.8813 * T)
        GMST0 = L0 + dL
        SIDTIME = GMST0 + lon

        # Calculate the moon's altitude and azimuth
        HA = SIDTIME - math.degrees(math.atan2(yh, xh))
        alt = math.asin(zh / r)
        az = math.degrees(
            math.atan2(math.sin(HA), math.cos(HA) * math.sin(lat) - math.tan(math.asin(zh / r)) * math.cos(lat)))
        if az < 0:
            az += 360

        return alt, az








import math
import datetime
import pytz


# start by calculating julian date
class Sun:
    def __init__(self):

        self.dayMs = 1000 * 60 * 60 * 24
        self.J1970 = 2440588
        self.J2000 = 2451545

    def toJulian(self, date):
        return date.timestamp() / self.dayMs - 0.5 + self.J1970

    def fromJulian(self, j):
        return datetime.datetime.fromtimestamp((j + 0.5 - self.J1970) * self.dayMs)

    def toDays(self):
        date = datetime.datetime.now()
        return self.toJulian(date) - self.J2000

    def get_julian_date(self):
        date = None
        # If no date is provided, use the current date and time
        if date is None:
            date = datetime.datetime.now()

        # Convert the date to a timestamp in milliseconds
        timestamp = date.timestamp() * 1000

        # Calculate the timezone offset in minutes
        if date.utcoffset():
            timezone_offset = date.utcoffset().total_seconds() // 60
        else:
            timezone_offset = 0

        # Calculate the Julian date and return it
        julian_date = (timestamp / 86400000) - (timezone_offset / 1440) + 2440587.5
        return julian_date

    def solardistance(self):
        juliandate = self.get_julian_date()
        # days since greenwich noon
        n = juliandate - 2451545
        # positions
        meanlong = 280.460 + 0.9856474 * n
        # g is mean anomaly
        g = 357.528 + 0.9856003 * n
        tmp = math.cos(g)
        temptwo = math.cos(2 * g)
        # solar distance is in astronomical units
        solardistance = 1.00014 - 0.01671 * tmp - 0.00014 * temptwo
        return solardistance * 92955807.3  # miles

    def solar_declination(self):
        # Get the current date (in UTC)
        now = datetime.datetime.utcnow()
        day_of_year = now.timetuple().tm_yday
        solar_declination = -23.45 * math.cos(math.radians((360 / 365) * (day_of_year + 10)))
        return solar_declination


    def observerAngle(self, height):
        return -2.076 * math.sqrt(height) / 60
    
    
    
    def eclipticlongitude(self, anomaly):
        tmp = math.pi / 180 * (
        1.9148 * math.sin(anomaly) + 0.02 * math.sin(2 * anomaly) + 0.0003 * math.sin(3 * anomaly))
        sectemp = math.pi / 180 * 102.9372
        return tmp + sectemp + math.pi



    def rightAscension(self, l, b):
        e = math.pi / 180 * 23.4397
        return math.atan2(math.sin(l) * math.cos(e) - math.tan(b) * math.sin(e), math.cos(l))

    def azimuth(self, h, phi, declination):
        return math.atan2(math.sin(h), math.cos(h) * math.sin(phi) - math.tan(declination) * math.cos(phi))
    
    def astro_info(self, year, month, day, hour, minute, second, lat, lon):
        # Convert latitude and longitude to radians
        lat_rad = math.radians(lat)
        lon_rad = math.radians(lon)
        ltr = []
        # Calculate the Julian Day (JD) for the given date and time
        jd = 367 * year - 7 * (year + (month + 9) // 12) // 4 + 275 * month // 9 + day + 1721013.5
        jd += (hour + minute / 60 + second / 3600) / 24
        ltr.append(jd)
        # jd, jc, geometric mean longitude, geometric mean anomaly, eccentricity of earth's orbit, equation of center, true longitude, true anomaly, sun's distance from earth (AU), longitude of omega, mean obliquity of ecliptic, sun's right ascension, sun's declination ,local hour angle.

        # Calculate the Julian Century (JC) for the given JD
        jc = (jd - 2451545) / 36525
        ltr.append(jc)

        # Calculate the Geometric Mean Longitude of the Sun (L0) in degrees
        l0 = 280.46646 + jc * (36000.76983 + jc * 0.0003032) % 360
        ltr.append(l0)

        # Calculate the Geometric Mean Anomaly of the Sun (M) in degrees
        m = 357.52911 + jc * (35999.05029 - 0.0001537 * jc)
        ltr.append(m)

        # Calculate the Eccentricity of Earth's Orbit (e)
        e = 0.016708634 - jc * (0.000042037 + 0.0000001267 * jc)
        ltr.append(e)

        # Calculate the Equation of Center (C) in degrees
        c = math.sin(math.radians(m)) * (1.914602 - jc * (0.004817 + 0.000014 * jc)) + math.sin(math.radians(2 * m)) * (0.019993 - 0.000101 * jc) + math.sin(math.radians(3 * m)) * 0.000289
        ltr.append(c)
        
        # Calculate the True Longitude of the Sun (l) in degrees
        l = l0 + c
        ltr.append(l)
        
        # Calculate the True Anomaly of the Sun (v) in degrees
        v = m + c
        ltr.append(v)
        
        # Calculate the Sun's Distance from Earth (r) in Astronomical Units (AU)
        r = 1.000001018 * (1 - e * e) / (1 + e * math.cos(math.radians(v)))
        ltr.append(r)

        # Calculate the Longitude of the Sun's Ascending Node (Omega) in degrees
        omega = 125.04 - 1934.136 * jc
        ltr.append(omega)

        # Calculate the Mean Obliquity of the Ecliptic (epsilon) in degrees
        epsilon = 23.439291 - jc * (0.0130042 + 0.00000016 * jc)
        ltr.append(epsilon)

        # Calculate the Sun's Right Ascension (alpha) in degrees
        alpha = math.degrees(math.atan2(math.cos(math.radians(epsilon)) * math.sin(math.radians(l)), math.cos(math.radians(l))))

        # Convert alpha to the range 0-360 degrees
        alpha = (alpha + 360) % 360
        ltr.append(alpha)

        # Calculate the Sun's Declination (delta) in degrees
        delta = math.degrees(math.asin(math.sin(math.radians(epsilon)) * math.sin(math.radians(l))))
        ltr.append(delta)

        # Calculate the Local Hour Angle (H) in degrees
        H = math.degrees(math.acos((math.sin(math.radians(-0.83)) - math.sin(math.radians(lat_rad)) * math.sin(math.radians(delta))) / (math.cos(math.radians(lat_rad)) * math.cos(math.radians(delta)))))
        ltr.append(H)
        
        return ltr


    def hourangle(self):
        # it is approximate (doesn't take into account minutes)
        if datetime.datetime.now().hour > 12:
            return 15 * (datetime.datetime.now().hour - 12)
        else:
            return 15 * (-(datetime.datetime.now().hour) + 12)
            
    def mean_solar_time(self, longitude):
        solartime = self.get_julian_date() - longitude/360
        return solartime # returns in approximate mean solar time
   
    def solar_mean_anomaly(self, longitude):
        solartime = self.mean_solar_time(longitude)
        anomaly = (357.5291 + 0.98560028 * solartime)%360
        return anomaly
    
    def center_equation(self, longitude):
        m = self.solar_mean_anomaly(longitude)
        c = 1.9148*math.sin(m) + 0.02 * math.sin(2 * m) + 0.0003*math.sin(3*m)
        return c
    
    def altitude(self, latitude, declination):
        return 90 - (latitude + declination) # returns in degrees
    
    def apparent_magnitude(self):
        # Apparent magnitude of Sun at a distance of 1 AU
        sun_magnitude_1au = -26.74
        # Calculate the apparent magnitude of the Sun at the current distance
        sun_magnitude = sun_magnitude_1au - 2.5 * math.log10(self.solardistance()/92955807.3) # to account for miles, we divide
        return sun_magnitude
    

    def sun_azimuth(self, lat, lon):
        # Get the current date and time in the observer's time zone
        now = datetime.datetime.now(pytz.timezone('US/Pacific'))

        # Convert the observer's latitude and longitude to radians
        lat = math.radians(lat)
        lon = math.radians(lon)

        # Calculate the observer's local sidereal time
        gst = now.astimezone(pytz.utc).strftime('%H:%M:%S')
        lst = (datetime.datetime.strptime(gst, '%H:%M:%S') + datetime.timedelta(hours=lon / math.pi * 12)).time()

        # Calculate the Sun's right ascension and declination for the current date and time
        d = now.date()
        n = (datetime.datetime(d.year, d.month, d.day, tzinfo=pytz.utc) - datetime.datetime(2000, 1, 1, 12,
                                                                                            tzinfo=pytz.utc)).days + 1
        g = 357.529 + 0.98560028 * n
        g = math.radians(g % 360)
        dec = math.asin(0.39779 * math.sin(g))
        ra = math.atan2(math.cos(g), 0.91746 * math.tan(dec))
        if ra < 0:
            ra += 2 * math.pi

        # Calculate the Sun's hour angle
        ha = math.radians((lst.hour + lst.minute / 60 + lst.second / 3600) * 15 - math.degrees(ra))

        # Calculate the Sun's azimuth using the formula
        sin_az = math.sin(ha) / (math.cos(lat) * math.tan(dec) - math.sin(lat) * math.cos(ha))
        cos_az = (math.sin(lat) * math.sin(dec) + math.cos(lat) * math.cos(dec) * math.cos(ha))
        azimuth = math.degrees(math.atan2(sin_az, cos_az))

        # Convert the azimuth to a compass direction
        if azimuth < 0:
            azimuth += 360
        if azimuth > 360:
            azimuth -= 360
        if azimuth < 11.25 or azimuth >= 348.75:
            return 'N'
        elif azimuth < 33.75:
            return 'NNE'
        elif azimuth < 56.25:
            return 'NE'
        elif azimuth < 78.75:
            return 'ENE'
        elif azimuth < 101.25:
            return 'E'
        elif azimuth < 123.75:
            return 'ESE'
        elif azimuth < 146.25:
            return 'SE'
        elif azimuth < 168.75:
            return 'SSE'
        elif azimuth < 191.25:
            return 'S'
        elif azimuth < 213.75:
            return 'SSW'
        elif azimuth < 236.25:
            return 'SW'
        elif azimuth < 258.75:
            return 'WSW'
        elif azimuth < 281.25:
            return 'W'
        elif azimuth < 303.75:
            return 'WNW'
        elif azimuth < 326.25:
            return 'NW'
        else:
            return 'NNW'





